# Python Standard Library Imports
from typing import Optional

# Other Third Party Imports
from pydantic import BaseModel


class ShippingDataDataClass(BaseModel):
    id: Optional[int]
    order_id: Optional[int]
    order: Optional[int]

    first_name: str
    last_name: str
    country: str
    state: str
    city: str
    street: str
    building: str
    floor: str
    apartment: str
    postal_code: str
    phone_number: str
    extra_description: str
    shipping_method: str
    email: str



class AbstractTestCase:
    """
    Abstract TestCase Class which provides a punch of an assertions utility functions
    """

    def assertEqual(self, actual, expected):
        assert actual == expected

    def assertIsNone(self, variable):
        assert variable is None

    def assertIsNotNone(self, variable):
        assert variable is not None

    def assertTrue(self, variable):
        assert variable is True

    def assertFalse(self, variable):
        assert variable is False

    def assertHasAttr(self, obj, attr):
        assert hasattr(obj, attr) is True

    def assertNoHasAttr(self, obj, attr):
        assert hasattr(obj, attr) is not True



# First Party Imports
from paymob.accept import AcceptUtils
from paymob.accept.config import URLsConfig
from tests.utils import AbstractTestCase


class TestAcceptUtils(AbstractTestCase):
    """Test AcceptUtils Class"""

    # =========== Start generate_merchant_order_id Tests =======
    # Test case to verify the behavior of the generate_merchant_order_id method.
    # It checks if the method returns the expected order ID based on the provided inputs.

    def test_generate_merchant_order_id_success(self):
        """test generate_merchant_order_id: Success"""

        mid_key = "Test"
        identifier = "1"

        expected_merchant_order_id = f"{mid_key}_{identifier}__000".split("__")[0]
        actual_merchant_order_id = AcceptUtils.generate_merchant_order_id(
            mid_key=mid_key,
            identifier=identifier,
        )
        self.assertEqual(actual_merchant_order_id.split("__")[0], expected_merchant_order_id)

    # ======= End generate_merchant_order_id Tests =======

    # ======= Start extract_mid_key_and_identifier Tests =======
    # Test case to verify the behavior of the extract_mid_key_and_identifier method.
    # It checks if the method correctly extracts the mid_key and identifier from a given merchant order id.

    def test_extract_mid_key_and_identifier_merchant_order_id_is_not_str(self):
        """test extract_mid_key_and_identifier: Mechant Order ID is not a String"""

        merchant_order_id = 1

        actual_mid_key, actual_identifier = AcceptUtils.extract_mid_key_and_identifier(
            merchant_order_id=merchant_order_id,
        )
        self.assertIsNone(actual_mid_key)
        self.assertIsNone(actual_identifier)

    def test_extract_mid_key_and_identifier_invalid_merchant_order_id(self):
        """test extract_mid_key_and_identifier: Invalid Mechant Order ID"""

        merchant_order_id = "Invalid"

        actual_mid_key, actual_identifier = AcceptUtils.extract_mid_key_and_identifier(
            merchant_order_id=merchant_order_id,
        )
        self.assertIsNone(actual_mid_key)
        self.assertIsNone(actual_identifier)

    def test_extract_mid_key_and_identifier_success(self):
        """test extract_mid_key_and_identifier:  Success"""
        mid_key = "Test"
        identifier = "1"
        merchant_order_id = f"{mid_key}_{identifier}__0"

        actual_mid_key, actual_identifier = AcceptUtils.extract_mid_key_and_identifier(
            merchant_order_id=merchant_order_id,
        )
        self.assertEqual(actual_mid_key, mid_key)
        self.assertEqual(actual_identifier, identifier)

    # ======= End extract_mid_key_and_identifier Tests =======

    # ======= Start create_iframe_url Tests ========
    # Test case to verify the behavior of the construct_iframe_url method.
    # It checks if the method constructs the correct iframe URL based on the provided iframe ID and payment key.

    def test_create_iframe_url_success(self):
        """test create_iframe_url: Success"""

        iframe_id = 112233
        payment_key = "*****"

        expected_iframe_url = URLsConfig.IFRAME.format(iframe_id=iframe_id, payment_key=payment_key)
        actual_iframe_url = AcceptUtils.create_iframe_url(iframe_id=iframe_id, payment_key=payment_key)
        self.assertEqual(actual_iframe_url, expected_iframe_url)

    # ======= End create_iframe_url Tests ========



# Python Standard Library Imports
from typing import Any, Dict, List, Tuple, Union

# Other Third Party Imports
from multimethod import overload

# First Party Imports
from paymob.data_classes import ResponseFeedBack

from .config import URLsConfig
from .connection import AcceptConnection
from .constants import PaymentSubTypes
from .factories import Invoice, Order, Product
from .response_codes import SUCCESS
from .transaction import Transaction


class AcceptAPIClient:
    """Class for Accept APIs
    By Initializing an Instance from This class, an auth token is obtained automatically
    and You will be able to call The Following APIs:
        1- Create Order
        2- Get Order
        3- Create Payment Key
        4- Pay
    """

    def __init__(self) -> None:
        # Initialize Connection
        self.connection = AcceptConnection()

    def create_order(
        self,
        merchant_order_id: str,
        amount_cents: str,
        currency: str,
        delivery_needed: bool = False,
        items: list = [],
        shipping_data: dict = {},
        shipping_details: dict = {},
    ) -> Tuple[str, Union[Order, None], ResponseFeedBack]:
        """Register an order to Accept's database

        Args:
            merchant_order_id (str): Internal Order ID
            amount_cents (str): Amount that will be paid (In Cents)
            currency (str): The currency related to this payment.
            delivery_needed (bool): Set it to be True if your order needs to be delivered by Accept's product delivery services. Defaults to False.
            items (list): list of objects contains the contents of the order if it is existing, send it as empty array if it is not available. Defaults to [].
            shipping_data (dict): Mandatory if your order needs to be delivered, otherwise you can delete the whole object. Defaults to {}.
            shipping_details (dict): Mandatory if your order needs to be delivered, otherwise you can delete the whole object. Defaults to {}.


        Returns:
            Tuple[str, Union[Order, None], ResponseFeedBack]: (Code, Order Insatnce, ResponseFeedBack Instance)
        """
        request_body = {
            "merchant_order_id": merchant_order_id,
            "amount_cents": str(amount_cents),
            "currency": currency,
            "delivery_needed": delivery_needed,
            "items": items,
            "shipping_data": shipping_data,
            "shipping_details": shipping_details,
        }

        code, order_data, feedback = self.connection.post(
            url=URLsConfig.CREATE_ORDER,
            json=request_body,
        )

        # TODO: Validates APIs Return Data Option
        order_instance = None
        if code == SUCCESS:
            order_instance = Order(**order_data)
            feedback.message = "Order with ID: {0} Created Successfully".format(order_instance.id)
        return code, order_instance, feedback

    def get_order(
        self,
        order_id: Union[int, str],
    ) -> Tuple[str, Union[Order, None], Union[str, None]]:
        """Retrives Order Data

        Args:
            order_id (Union[int, str]): Paymob's External Order ID

        Returns:
            Tuple[str, Union[Order, None], ResponseFeedBack]: (Code, Order Instance, ResponseFeedBack Instance)
        """

        code, order_data, feedback = self.connection.get(
            url=URLsConfig.GET_ORDER.format(order_id=order_id),
        )

        # TODO: Validates APIs Return Data Option
        order_instance = None
        if code == SUCCESS:
            order_instance = Order(**order_data)
            feedback.message = "Successfully Retrieved Order: {0}".format(order_id)
        return code, order_instance, feedback

    def create_payment_key(
        self,
        order_id: str,
        amount_cents: str,
        currency: str,
        billing_data: dict,
        integration_id: int,
        card_token_key: str = None,
        expiration: int = 3600,  # 1 Hour
        lock_order_when_paid: bool = False,
    ) -> Tuple[str, str, ResponseFeedBack]:
        """Create Payment Key

        Args:
            order_id (str): External order id
            expiration (int): The expiration time of this payment token in seconds. (The maximum is 3600 seconds which is an hour). Defaulkts to 3600.
            amount_cents (str): Amount that will be paid
            currency (str): Payment Currency
            billing_data (dict): The billing data related to the customer related to this payment.
            integration_id (int): Paymob's Integration ID
            card_token_key (str, optional): Card token. Defaults to None.
            lock_order_when_paid (bool, optional): A flag prevent this order to be paid again if it is paid. Defaults to False.

        Returns:
            Tuple[str, str, ResponseFeedBack]: (Code, Payment Key, ResponseFeedBack Instance)
        """

        request_body = {
            "order_id": order_id,
            "amount_cents": str(amount_cents),
            "currency": str(currency),
            "expiration": expiration,
            "integration_id": integration_id,
            "billing_data": billing_data,
            "lock_order_when_paid": lock_order_when_paid,
        }
        if card_token_key:
            request_body["token"] = card_token_key

        code, data, feedback = self.connection.post(
            url=URLsConfig.PAYMENT_KEY,
            json=request_body,
        )

        # TODO: Validates APIs Return Data Option
        payment_key = None
        if code == SUCCESS:
            feedback.message = "Payment Key for Order: {0} Created Successfully".format(order_id)
            payment_key = data.get("token")
        return code, payment_key, feedback

    def proceed_kiosk_payment(
        self,
        payment_key: str,
    ) -> Tuple[str, Union[Transaction, None], ResponseFeedBack]:
        """Proceed Kiosk Payment

        Args:
            payment_key (str): Obtained Payment Key

        Returns:
            Tuple[str, Union[Transaction, None], ResponseFeedBack]: (Code, Payment Data Dict, ResponseFeedBack Instance)
        """

        request_body = {
            "source": {
                "identifier": PaymentSubTypes.AGGREGATOR,
                "subtype": PaymentSubTypes.AGGREGATOR,
            },
            "payment_token": payment_key,
        }

        code, transaction_data, feedback = self.connection.post(
            url=URLsConfig.PAY,
            json=request_body,
        )

        # TODO: Validates APIs Return Data Option
        transaction_instance = None
        if code == SUCCESS:
            transaction_instance = Transaction(connection=self.connection, **transaction_data)
            feedback.message = "Kiosk Payment Processed Successfully"
        return code, transaction_instance, feedback

    def proceed_wallet_payment(
        self,
        payment_key: str,
        identifier: str,
    ) -> Tuple[str, Union[Transaction, None], ResponseFeedBack]:
        """Proceed Wallet Payment

        Args:
            payment_key (str): Obtained Payment Key
            identifier (str): Wallet Number or AGGREGATOR for Kiosk Payment

        Returns:
            Tuple[str, Union[Transaction, None], ResponseFeedBack]: (Code, Payment Data Dict, ResponseFeedBack Instance)
        """

        request_body = {
            "source": {
                "identifier": identifier,
                "subtype": PaymentSubTypes.WALLET,
            },
            "payment_token": payment_key,
        }

        code, transaction_data, feedback = self.connection.post(
            url=URLsConfig.PAY,
            json=request_body,
        )

        # TODO: Validates APIs Return Data Option
        transaction_instance = None
        if code == SUCCESS:
            transaction_instance = Transaction(connection=self.connection, **transaction_data)
            feedback.message = "Wallet Payment for Identifier: {0} Processed Successfully".format(identifier)
        return code, transaction_instance, feedback

    def proceed_cash_payment(
        self,
        payment_key: str,
    ) -> Tuple[str, Union[Transaction, None], ResponseFeedBack]:
        """Proceed Cash Payment

        Args:
            payment_key (str): Obtained Payment Key

        Returns:
            Tuple[str, Union[Transaction, None], ResponseFeedBack]: (Code, Payment Data Dict, ResponseFeedBack Instance)
        """

        request_body = {
            "source": {
                "identifier": PaymentSubTypes.CASH.lower(),
                "subtype": PaymentSubTypes.CASH,
            },
            "payment_token": payment_key,
        }

        code, transaction_data, feedback = self.connection.post(
            url=URLsConfig.PAY,
            json=request_body,
        )

        # TODO: Validates APIs Return Data Option
        transaction_instance = None
        if code == SUCCESS:
            transaction_instance = Transaction(connection=self.connection, **transaction_data)
            feedback.message = "Cash Payment Processed Successfully"
        return code, transaction_instance, feedback

    def proceed_card_token_payment(
        self,
        payment_key: str,
        card_token: str,
    ) -> Tuple[str, Union[Transaction, None], ResponseFeedBack]:
        """Proceed With Saved Card Payment

        Args:
            payment_key (str): Obtained Payment Key
            card_token (str): Saved Card Token

        Returns:
            Tuple[str, Union[Transaction, None], ResponseFeedBack]: (Code, Payment Data Dict, ResponseFeedBack Instance)
        """

        request_body = {
            "source": {
                "identifier": card_token,
                "subtype": PaymentSubTypes.TOKEN,
            },
            "payment_token": payment_key,
        }

        code, transaction_data, feedback = self.connection.post(
            url=URLsConfig.PAY,
            json=request_body,
        )

        # TODO: Validates APIs Return Data Option
        transaction_instance = None
        if code == SUCCESS:
            transaction_instance = Transaction(connection=self.connection, **transaction_data)
            feedback.message = "Payment With Saved Card Token Processed Successfully"
        return code, transaction_instance, feedback

    def create_invoice_link(
        self,
        amount_cents: str,
        shipping_data: Dict[str, Any],
        items: List[Dict[str, Any]],
        currency: str,
        integrations: List[int],
        delivery_needed: bool = False,
    ) -> Tuple[str, Union[Invoice, None], ResponseFeedBack]:
        """_summary_

        Args:
            amount_cents (str): The amount in cents to be paid for in the invoice.
            shipping_data (Dict[str, Any]): The details of the customer or end-user.
            items (List[Dict[str, Any]]): It will include the details of the order.
            currency (str): The currency used in the invoice. By default, it will be EGP.
            integrations (List[int]): The payment methods that will be listed in the invoice link. You should enter the integration ID of every payment method.
            delivery_needed (bool): Determines if you use our delivery. Defaults to False.

        Returns:
            Tuple[str, Union[Invoice, None], ResponseFeedBack]: (Code, Invoice Instance, ResponseFeedBack Instance)
        """

        request_body = {
            "api_source": "INVOICE",
            "amount_cents": amount_cents,
            "currency": currency,
            "shipping_data": shipping_data,
            "integrations": integrations,
            "items": items,
            "delivery_needed": delivery_needed,
        }

        code, invoice_data, feedback = self.connection.post(
            url=URLsConfig.CREATE_INVOICE_LINK,
            json=request_body,
        )

        # TODO: Validates APIs Return Data Option
        invoice_instance = None
        if code == SUCCESS:
            invoice_instance = Invoice(**invoice_data)
            feedback.message = "Invoice Link Data Retrieved Successfully"
        return code, invoice_instance, feedback

    def create_product_link(
        self,
        product_name: str,
        product_description: str,
        amount_cents: str,
        currency: str,
        inventory: str,
        integrations: List[int],
        allow_quantity_edit: bool = False,
        delivery_needed: bool = False,
    ) -> Tuple[str, Union[Product, None], ResponseFeedBack]:
        """Creates a Product Link

        Args:
            product_name (str): The name of your product.
            product_description (str): The description of your product.
            amount_cents (str): The price in cents of the product.
            currency (str): The currency used in the invoice. By default, it will be EGP.
            inventory (str): The stock of your product in your inventory.
            integrations (List[int]): The payment methods that will be listed in the invoice link. You should enter the integration ID of every payment method.
            allow_quantity_edit (bool): If the stock will be reduced when a product is paid for. Defaults to False.
            delivery_needed (bool): Determines if you use our delivery. Defaults to False.

        Returns:
            Tuple[str, Union[Product, None], ResponseFeedBack]: (Code, Product Instance, ResponseFeedBack Instance)
        """

        request_body = {
            "product_name": product_name,
            "product_description": product_description,
            "amount_cents": amount_cents,
            "currency": currency,
            "inventory": inventory,
            "integrations": integrations,
            "allow_quantity_edit": allow_quantity_edit,
            "delivery_needed": delivery_needed,
        }

        code, product_data, feedback = self.connection.post(
            url=URLsConfig.CREATE_PRODUCT_LINK,
            json=request_body,
        )

        # TODO: Validates APIs Return Data Option
        product_instance = None
        if code == SUCCESS:
            product_instance = Product(**product_data)
            feedback.message = "Product Link Data Retrieved Successfully"
        return code, product_instance, feedback

    @overload
    def get_transaction(
        self,
        transaction_id: int,
    ) -> Tuple[str, Union[Transaction, None], ResponseFeedBack]:
        """Retrieves Transaction Data by Transaction ID

        Args:
            transaction_id (int): Paymob's Transaction ID

        Returns:
            Tuple[str, Union[Transaction, None], ResponseFeedBack]: (Code, Transaction Instance, ResponseFeedBack Instance)
        """

        code, transaction_data, feedback = self.connection.get(
            url=URLsConfig.GET_TRANSACTION_ID.format(transaction_id=transaction_id),
        )
        # TODO: Validates APIs Return Data Option
        transaction_instance = None
        if code == SUCCESS:
            transaction_instance = Transaction(connection=self.connection, **transaction_data)
            feedback.message = "Transaction: {0} Retrieved Successfully".format(transaction_instance.id)
        return code, transaction_instance, feedback

    @overload
    def get_transaction(
        self,
        merchant_order_id: str = None,
        order_id: int = None,
    ) -> Tuple[str, Union[Transaction, None], ResponseFeedBack]:
        """Retrieves Transaction Data by Merchant Order ID and Order ID

        Args:
            merchant_order_id (str): Internal Order ID. Defaults to None.
            order_id (int): Paymob's External Order ID. Defaults to None.

        Note: Either merchant order_id or order_id must be passed

        Returns:
            Tuple[str, Union[Transaction, None], ResponseFeedBack]: (Code, Transaction Instance, ResponseFeedBack Instance)
        """

        request_body = {}
        if merchant_order_id:
            request_body["merchant_order_id"] = merchant_order_id
        if order_id:
            request_body["order_id"] = order_id

        code, transaction_data, feedback = self.connection.post(
            url=URLsConfig.GET_TRANSACTION_MERCHANT_ID,
            json=request_body,
        )

        # TODO: Validates APIs Return Data Option
        transaction_instance = None
        if code == SUCCESS:
            transaction_instance = Transaction(connection=self.connection, **transaction_data)
            feedback.message = (
                "Transaction: {0} of Order ID: {1} and Merchant Order ID: {2} Retrieved Successfully".format(
                    transaction_instance.id,
                    order_id,
                    merchant_order_id,
                )
            )
        return code, transaction_instance, feedback



# Python Standard Library Imports


class BaseClass(object):
    def __init__(self, classtype):
        self._type = classtype

    def __str__(self) -> str:
        if hasattr(self, "id"):
            return f"{self.__class__.__name__} No: {self.id}"
        return self.__repr__()


def ClassFactory(name, BaseClass=BaseClass):
    splitted_name = name.split("_")
    if len(splitted_name) > 1:
        name = "".join(element.capitalize() for element in splitted_name)

    def _wrap(key, value):
        if isinstance(value, (tuple, list, set, frozenset)):
            return type(value)([_wrap(key, v) for v in value])
        else:
            klass = ClassFactory(key)
            return klass(**value) if isinstance(value, dict) else value

    def __init__(self, **kwargs):

        for key, value in kwargs.items():
            # here, the argnames variable is the one passed to the
            # ClassFactory call
            setattr(self, key, _wrap(key, value))

        BaseClass.__init__(self, name[: -len("Class")])

    newclass = type(name, (BaseClass,), {"__init__": __init__})
    return newclass


__all__ = [ClassFactory]



# Python Standard Library Imports
from unittest.mock import patch

# First Party Imports
from paymob.accept import HMACValidator
from paymob.accept.constants import AcceptCallbackTypes
from tests.utils import AbstractTestCase


class TestHMACValidator(AbstractTestCase):
    """Test HMACValidator Class"""

    # ======== Start SetUpClass ========

    @classmethod
    def setup_class(cls):

        cls.valid_incoming_hmac = "*****"
        cls.transaction_callback_dict = {
            "type": AcceptCallbackTypes.TRANSACTION,
            "obj": {
                "id": 109628234,
                "pending": False,
                "amount_cents": 1000,
                "success": True,
                "is_auth": False,
                "is_capture": False,
                "is_standalone_payment": True,
                "is_voided": False,
                "is_refunded": False,
                "is_3d_secure": False,
                "integration_id": 3871725,
                "profile_id": 787951,
                "has_parent_transaction": False,
                "order": {
                    "id": 126087094,
                    "created_at": "2023-06-01T21:45:56.993153",
                    "delivery_needed": False,
                    "merchant": {
                        "id": 787951,
                        "created_at": "2023-05-14T19:41:48.955870",
                        "phones": ["+201028362012"],
                        "company_emails": [],
                        "company_name": "Mahmoud",
                        "state": "",
                        "country": "EGY",
                        "city": "Cairo",
                        "postal_code": "",
                        "street": "",
                    },
                    "collector": None,
                    "amount_cents": 1000,
                    "shipping_data": {
                        "id": 61920183,
                        "first_name": "Brett",
                        "last_name": "Roberts",
                        "street": "NA",
                        "building": "NA",
                        "floor": "NA",
                        "apartment": "NA",
                        "city": "NA",
                        "state": "NA",
                        "country": "NA",
                        "email": "swsz2qbisz@amenli.com",
                        "phone_number": "01550272164",
                        "postal_code": "NA",
                        "extra_description": "",
                        "shipping_method": "UNK",
                        "order_id": 126087094,
                        "order": 126087094,
                    },
                    "currency": "EGP",
                    "is_payment_locked": False,
                    "is_return": False,
                    "is_cancel": False,
                    "is_returned": False,
                    "is_canceled": False,
                    "merchant_order_id": "lead_20__1685645156",
                    "wallet_notification": None,
                    "paid_amount_cents": 0,
                    "notify_user_with_email": False,
                    "items": [],
                    "order_url": "https://accept.paymob.com/standalone/?ref=i_LRR2Q3VHeUtzaTBFb0xSdG0rRVY2YzA4dz09X0JLelBTZElyRXRpM3hkR0tHU1RDQlE9PQ",
                    "commission_fees": 0,
                    "delivery_fees_cents": 0,
                    "delivery_vat_cents": 0,
                    "payment_method": "tbc",
                    "merchant_staff_tag": None,
                    "api_source": "OTHER",
                    "data": {},
                },
                "created_at": "2023-06-01T21:45:57.830530",
                "transaction_processed_callback_responses": [
                    {
                        "response": {
                            "status": "200",
                            "content": "None",
                            "headers": "{'Server': 'nginx', 'Content-Type': 'text/plain; charset=UTF-8', 'Transfer-Encoding': 'chunked', 'Vary': 'Accept-Encoding', 'X-Request-Id': 'd1cde7b7-5fda-4ea4-9d3b-ddc9b12ff94e', 'X-Token-Id': 'ac84f01e-0bc5-41d3-be08-b5b3912f8148', 'Cache-Control': 'no-cache, private', 'Date': 'Thu, 01 Jun 2023 18:46:17 GMT', 'Content-Encoding': 'gzip'}",
                            "encoding": "UTF-8",
                        },
                        "callback_url": "https://webhook.site/ac84f01e-0bc5-41d3-be08-b5b3912f8148",
                        "response_received_at": "2023-06-01T18:46:17.484494",
                    },
                ],
                "currency": "EGP",
                "source_data": {
                    "pan": "01010101010",
                    "type": "wallet",
                    "sub_type": "wallet",
                    "owner_name": None,
                    "phone_number": "01010101010",
                },
                "api_source": "OTHER",
                "terminal_id": None,
                "merchant_commission": 0,
                "installment": None,
                "discount_details": [],
                "is_void": False,
                "is_refund": False,
                "data": {
                    "klass": "WalletPayment",
                    "token": "",
                    "amount": 1000,
                    "method": 0,
                    "message": "Transaction has been completed successfully.",
                    "currency": "EGP",
                    "created_at": "2023-06-01T18:45:57.876617",
                    "mpg_txn_id": "213412341",
                    "order_info": "swsz2qbisz@amenli.com",
                    "uig_txn_id": "4351324134",
                    "upg_txn_id": None,
                    "mer_txn_ref": "3871725_3fb960d3f7495a998f8984e89b978f10",
                    "redirect_url": "https://accept.paymobsolutions.com/api/acceptance/wallet_other_test/wallet_template?token=****",
                    "wallet_issuer": "VODAFONE",
                    "wallet_msisdn": "01010101010",
                    "gateway_source": "",
                    "upg_qrcode_ref": "4351324134",
                    "txn_response_code": "200",
                    "gateway_integration_pk": 3871725,
                },
                "is_hidden": False,
                "payment_key_claims": {
                    "extra": {},
                    "pmk_ip": "197.62.181.81",
                    "user_id": 1363533,
                    "currency": "EGP",
                    "order_id": 126087094,
                    "amount_cents": 1000,
                    "billing_data": {
                        "city": "NA",
                        "email": "swsz2qbisz@amenli.com",
                        "floor": "NA",
                        "state": "NA",
                        "street": "NA",
                        "country": "NA",
                        "building": "NA",
                        "apartment": "NA",
                        "last_name": "Roberts",
                        "first_name": "Brett",
                        "postal_code": "NA",
                        "phone_number": "01550272164",
                        "extra_description": "NA",
                    },
                    "integration_id": 3871725,
                    "lock_order_when_paid": False,
                    "single_payment_attempt": False,
                },
                "error_occured": False,
                "is_live": False,
                "other_endpoint_reference": "4351324134",
                "refunded_amount_cents": 0,
                "source_id": -1,
                "is_captured": False,
                "captured_amount": 0,
                "merchant_staff_tag": None,
                "updated_at": "2023-06-17T17:53:32.176405",
                "is_settled": False,
                "bill_balanced": False,
                "is_bill": False,
                "owner": 1363533,
                "parent_transaction": None,
            },
            "transaction_processed_callback_responses": "",
        }
        cls.card_token_callback_dict = {
            "type": AcceptCallbackTypes.CARD_TOKEN,
            "obj": cls.transaction_callback_dict.get("obj"),
        }
        cls.delivery_status_callback_dict = {
            "type": AcceptCallbackTypes.DELIVERY_STATUS,
            "obj": cls.transaction_callback_dict.get("obj"),
        }
        cls.invalid_transaction_dict = {
            "type": "Invalid",
            "obj": cls.transaction_callback_dict.get("obj"),
        }

    # ======== End SetUpClass ========

    # ========= Start __init__ Tests =========

    def test___init__callback_dict_is_not_dict(self):
        """test __init__: callback_dict is not of type dict"""

        callback_dict = "invalid"
        hmac = HMACValidator(
            incoming_hmac=self.valid_incoming_hmac,
            callback_dict=callback_dict,
        )
        self.assertNoHasAttr(hmac, "callback_obj_dict")
        self.assertEqual(hmac.callback_dict, callback_dict)

    def test___init__success(self):
        """test __init__: Success"""

        callback_obj_dict = {"id": 1}
        callback_dict = {"obj": callback_obj_dict}
        hmac = HMACValidator(
            incoming_hmac=self.valid_incoming_hmac,
            callback_dict=callback_dict,
        )
        self.assertEqual(hmac.callback_obj_dict, callback_obj_dict)
        self.assertEqual(hmac.callback_dict, callback_dict)

    # ========= End __init__ Tests =========

    # ========= Start is_valid Tests =========
    # Test cases to verify the behavior of the is_valid method for a Transaction/CardToken/DeliveryStatus callbacks.

    def test_is_valid_callback_is_not_a_dict(self):
        """test is_valid: Callback is not a Dict"""
        hmac = HMACValidator(
            incoming_hmac=self.valid_incoming_hmac,
            callback_dict="invalid",
        )
        self.assertFalse(hmac.is_valid)

    def test_is_valid_invalid_callback_type(self):
        """test is_valid: Invalid callback_type"""
        hmac = HMACValidator(
            incoming_hmac=self.valid_incoming_hmac,
            callback_dict=self.invalid_transaction_dict,
        )
        self.assertFalse(hmac.is_valid)

    @patch.object(HMACValidator, "_generate_transaction_processed_hmac")
    def test_is_valid_invalid_incoming_hmac(self, mock_generate_transaction_processed_hmac):
        """test is_valid: Invalid Incoming HMAC"""
        mock_generate_transaction_processed_hmac.return_value = "Invalid"
        hmac = HMACValidator(
            incoming_hmac=self.valid_incoming_hmac,
            callback_dict=self.transaction_callback_dict,
        )
        self.assertFalse(hmac.is_valid)
        self.assertTrue(mock_generate_transaction_processed_hmac.called)

    @patch.object(HMACValidator, "_generate_transaction_processed_hmac")
    def test_is_valid_transaction_callback_success(self, mock_generate_transaction_processed_hmac):
        """test is_valid: callback_type=TRANSACTION Success"""

        mock_generate_transaction_processed_hmac.return_value = self.valid_incoming_hmac
        hmac = HMACValidator(
            incoming_hmac=self.valid_incoming_hmac,
            callback_dict=self.transaction_callback_dict,
        )
        self.assertTrue(hmac.is_valid)
        self.assertTrue(mock_generate_transaction_processed_hmac.called)

    @patch.object(HMACValidator, "_generate_card_token_processed_hmac")
    def test_is_valid_card_token_callback_success(self, mock_generate_card_token_processed_hmac):
        """test is_valid: callback_type=CARD_TOKEN Success"""

        mock_generate_card_token_processed_hmac.return_value = self.valid_incoming_hmac
        hmac = HMACValidator(
            incoming_hmac=self.valid_incoming_hmac,
            callback_dict=self.card_token_callback_dict,
        )
        self.assertTrue(hmac.is_valid)
        self.assertTrue(mock_generate_card_token_processed_hmac.called)

    @patch.object(HMACValidator, "_generate_delivery_status_processed_hmac")
    def test_is_valid_delivery_status_callback_success(self, mock_generate_delivery_status_processed_hmac):
        """test is_valid: callback_type=DELIVERY_STATUS Success"""

        mock_generate_delivery_status_processed_hmac.return_value = self.valid_incoming_hmac
        hmac = HMACValidator(
            incoming_hmac=self.valid_incoming_hmac,
            callback_dict=self.delivery_status_callback_dict,
        )
        self.assertTrue(hmac.is_valid)
        self.assertTrue(mock_generate_delivery_status_processed_hmac.called)

    # ========= End is_valid Tests =========

    # ========= Start _generate_processed_hmac Tests =========

    @patch.object(HMACValidator, "_calculate_hmac")
    def test_generate_processed_hmac_hmac_dict_is_not_dict(self, mock_calculate_hmac):
        """test _generate_processed_hmac: hmac_dict is not of type dict"""

        hmac = HMACValidator(
            incoming_hmac=self.valid_incoming_hmac,
            callback_dict=self.delivery_status_callback_dict,
        )
        actual_hmac_message = hmac._generate_processed_hmac(
            hmac_dict="Invalid",
        )
        expected_hmac_message = ""
        self.assertEqual(actual_hmac_message, expected_hmac_message)
        self.assertFalse(mock_calculate_hmac.called)

    @patch.object(HMACValidator, "_calculate_hmac")
    def test_generate_processed_hmac_success(self, mock_calculate_hmac):
        """test _generate_processed_hmac: Success"""

        expected_hmac_message = "HMAC Message"
        mock_calculate_hmac.return_value = expected_hmac_message

        hmac = HMACValidator(
            incoming_hmac=self.valid_incoming_hmac,
            callback_dict=self.delivery_status_callback_dict,
        )
        actual_hmac_message = hmac._generate_processed_hmac(
            hmac_dict={},
        )
        self.assertEqual(actual_hmac_message, expected_hmac_message)
        self.assertTrue(mock_calculate_hmac.called)

    # ========= End _generate_processed_hmac Tests =========

    # ========= Start _generate_transaction_processed_hmac Tests =========

    @patch.object(HMACValidator, "_generate_processed_hmac")
    def test_generate_transaction_processed_hmac_invalid_callback_obj_dict(self, mock_generate_processed_hmac):
        """test _generate_transaction_processed_hmac: Invalid callback_obj_dict"""

        hmac = HMACValidator(
            incoming_hmac=self.valid_incoming_hmac,
            callback_dict={},
        )
        actual_hmac = hmac._generate_transaction_processed_hmac()
        expected_hmac = ""
        self.assertEqual(actual_hmac, expected_hmac)
        self.assertFalse(mock_generate_processed_hmac.called)

    @patch.object(HMACValidator, "_generate_processed_hmac")
    def test_generate_transaction_processed_hmac_success(self, mock_generate_processed_hmac):
        """test _generate_transaction_processed_hmac: Success"""

        expected_hmac = "*****"
        mock_generate_processed_hmac.return_value = expected_hmac

        hmac = HMACValidator(
            incoming_hmac=self.valid_incoming_hmac,
            callback_dict=self.transaction_callback_dict,
        )
        actual_hmac = hmac._generate_transaction_processed_hmac()
        self.assertEqual(actual_hmac, expected_hmac)
        self.assertTrue(mock_generate_processed_hmac.called)

    # ========= End _generate_transaction_processed_hmac Tests =========

    # ========= Start _generate_card_token_processed_hmac Tests =========

    @patch.object(HMACValidator, "_generate_processed_hmac")
    def test_generate_card_token_processed_hmac_invalid_callback_obj_dict(self, mock_generate_processed_hmac):
        """test _generate_card_token_processed_hmac: Invalid callback_obj_dict"""

        hmac = HMACValidator(
            incoming_hmac=self.valid_incoming_hmac,
            callback_dict={},
        )
        actual_hmac = hmac._generate_card_token_processed_hmac()
        expected_hmac = ""
        self.assertEqual(actual_hmac, expected_hmac)
        self.assertFalse(mock_generate_processed_hmac.called)

    @patch.object(HMACValidator, "_generate_processed_hmac")
    def test_generate_card_token_processed_hmac_hmac_success(self, mock_generate_processed_hmac):
        """test _generate_card_token_processed_hmac: Success"""

        expected_hmac = "*****"
        mock_generate_processed_hmac.return_value = expected_hmac

        hmac = HMACValidator(
            incoming_hmac=self.valid_incoming_hmac,
            callback_dict=self.card_token_callback_dict,
        )
        actual_hmac = hmac._generate_card_token_processed_hmac()
        self.assertEqual(actual_hmac, expected_hmac)
        self.assertTrue(mock_generate_processed_hmac.called)

    # ========= End _generate_card_token_processed_hmac Tests =========

    # ========= Start _generate_card_token_processed_hmac Tests =========

    @patch.object(HMACValidator, "_generate_processed_hmac")
    def test_generate_delivery_status_processed_hmac_invalid_callback_obj_dict(self, mock_generate_processed_hmac):
        """test _generate_delivery_status_processed_hmac: Invalid callback_obj_dict"""

        hmac = HMACValidator(
            incoming_hmac=self.valid_incoming_hmac,
            callback_dict={},
        )
        actual_hmac = hmac._generate_delivery_status_processed_hmac()
        expected_hmac = ""
        self.assertEqual(actual_hmac, expected_hmac)
        self.assertFalse(mock_generate_processed_hmac.called)

    @patch.object(HMACValidator, "_generate_processed_hmac")
    def test_generate_delivery_status_processed_hmac_hmac_success(self, mock_generate_processed_hmac):
        """test _generate_delivery_status_processed_hmac: Success"""

        expected_hmac = "*****"
        mock_generate_processed_hmac.return_value = expected_hmac

        hmac = HMACValidator(
            incoming_hmac=self.valid_incoming_hmac,
            callback_dict=self.card_token_callback_dict,
        )
        actual_hmac = hmac._generate_delivery_status_processed_hmac()
        self.assertEqual(actual_hmac, expected_hmac)
        self.assertTrue(mock_generate_processed_hmac.called)

    # ========= End _generate_delivery_status_processed_hmac Tests =========



# First Party Imports
from paymob.utils import ClassFactory

DynamicTransactionFactory = ClassFactory("Transaction")
Order = ClassFactory("Order")
Invoice = ClassFactory("Invoice")
Product = ClassFactory("Product")
DeliveryStatusCallbackFactory = ClassFactory("DeliveryStatusCallback")
CardTokenCallbackFactory = ClassFactory("CardTokenCallback")



# Python Standard Library Imports
from typing import Union

# Other Third Party Imports
from pydantic import BaseModel

from .shipping_data import ShippingDataDataClass


class PaymentKeyClaimsDataClass(BaseModel):
    exp: Union[int, float]
    currency: str
    amount_cents: int
    user_id: int
    integration_id: int
    shipping_data: ShippingDataDataClass



# First Party Imports
from paymob.accept.factories import CardTokenCallbackFactory
from paymob.accept.hmac_validator import HMACValidator


class CardTokenCallback(HMACValidator, CardTokenCallbackFactory):
    pass


__all__ = ["CardTokenCallback"]



# First Party Imports
from paymob.accept.factories import DynamicTransactionFactory
from paymob.accept.hmac_validator import HMACValidator


class TransactionCallback(HMACValidator, DynamicTransactionFactory):
    pass


__all__ = ["TransactionCallback"]



class AcceptCallbackTypes:
    TRANSACTION = "TRANSACTION"
    CARD_TOKEN = "TOKEN"
    DELIVERY_STATUS = "DELIVERY_STATUS"


class PaymentSubTypes:
    AGGREGATOR = "AGGREGATOR"
    WALLET = "WALLET"
    CASH = "CASH"
    TOKEN = "TOKEN"


class DeliveryStatuses:
    SCHEDUALED = "Scheduled"
    CONTACTING_MERCHANT = "Contacting Merchant"
    PICKING_UP = "Picking Up"
    COURIER_RECEIVED = "Courier Received"
    AT_WAREHOUSE = "At Warehouse"
    AGENT_OUT = "Agent Out"
    ON_ROUTE = "On Route"
    AT_CUSTOMER = "At Customer"
    DELIVERED = "Delivered"
    CANCELED = "Canceled"
    DELIVERY_FAILED = "Delivery Failed"
    RETURN_SCHEDULED = "Return Scheduled"
    PACKAGE_RETURNED = "Package Returned"



# Python Standard Library Imports
import time
from typing import Tuple, Union

from .config import URLsConfig


class AcceptUtils:
    @staticmethod
    def generate_merchant_order_id(mid_key: str, identifier: int):
        """Constructs merchant order id from mid_key and identifier

        Args:
            mid_key (str):
            identifier (int): instance id

        Returns:
            str: Internal Order ID (Format: <mid_key>_<identifier>__<time>)
        """
        return "{0}_{1}__{2}".format(mid_key, identifier, int(time.time()))

    @staticmethod
    def extract_mid_key_and_identifier(merchant_order_id: str) -> Tuple[Union[str, None], Union[str, None]]:
        """Extracts mid_key and identifier from merchant order id

        Args:
            merchant_order_id (str): Internal Order ID (Format: <mid_key>_<identifier>__<time>

        Returns
            Tuple[Union[str, None], Union[str, None]]: mid_key, identifier
        """

        if not isinstance(merchant_order_id, str):
            return None, None

        mid_raw = merchant_order_id.split("__")[0]
        mid_ref_list = mid_raw.split("_")
        if len(mid_ref_list) != 2:
            return None, None

        mid_key = mid_ref_list[0]
        identifier = mid_ref_list[1]
        return mid_key, identifier

    @staticmethod
    def create_iframe_url(iframe_id: int, payment_key: str) -> str:
        """Constructs Iframe URL

        Args:
            iframe_id (int): Paymob's Iframe ID
            payment_key (str): Generated Payment Key

        Returns:
            str: Full Iframe URL
        """
        return URLsConfig.IFRAME.format(
            iframe_id=iframe_id,
            payment_key=payment_key,
        )



# Python Standard Library Imports
import hashlib
import hmac
from typing import Any, Dict

from .config import Credentials
from .constants import AcceptCallbackTypes


class HMACValidator:
    def __init__(self, incoming_hmac: str, callback_dict: Dict[str, Any], **kwargs) -> None:
        """Initialize HMAC Attributes

        Args:
            incoming_hmac (str): Incoming Paymob's HMAC
            callback_dict Dict[str, Any]: Incoming Callback Dict
        """
        self.incoming_hmac = incoming_hmac
        self.callback_dict = callback_dict
        if isinstance(self.callback_dict, dict):
            self.callback_obj_dict = self.callback_dict.get("obj")

        super().__init__(**kwargs)

    @staticmethod
    def _calculate_hmac(message: str) -> str:
        """Calculates HMAC

        Args:
            message (str): GeneratedHMAC Message

        Returns:
            str: Calculated HMAC
        """

        return (
            hmac.new(
                Credentials.ACCEPT_HMAC_SECRET.encode("utf-8"),
                message.encode("utf-8"),
                hashlib.sha512,
            )
            .hexdigest()
            .lower()
        )

    @classmethod
    def _generate_processed_hmac(cls, hmac_dict: Dict[str, Any]) -> str:
        """Creates HMAC from sent self.callback_obj_dict

        Args:
            hmac_dict (Dict[str, Any]): Hmac Dict

        Returns:
            str: Generated HMAC
        """
        if not isinstance(hmac_dict, dict):
            return ""

        message = ""
        for value in hmac_dict.values():
            if isinstance(value, bool):
                value = str(value).lower()
            if value is None:
                value = ""
            message += str(value)

        return cls._calculate_hmac(message=message)

    def _generate_transaction_processed_hmac(self) -> str:
        """Creates HMAC from sent transaction callback self.callback_obj_dict

        Returns:
            str: Generated HMAC
        """
        if not isinstance(self.callback_obj_dict, dict):
            return ""

        hmac_dict = {
            "amount_cents": self.callback_obj_dict.get("amount_cents"),
            "created_at": self.callback_obj_dict.get("created_at"),
            "currency": self.callback_obj_dict.get("currency"),
            "error_occured": self.callback_obj_dict.get("error_occured"),
            "has_parent_transaction": self.callback_obj_dict.get("has_parent_transaction"),
            "id": self.callback_obj_dict.get("id"),
            "integration_id": self.callback_obj_dict.get("integration_id"),
            "is_3d_secure": self.callback_obj_dict.get("is_3d_secure"),
            "is_auth": self.callback_obj_dict.get("is_auth"),
            "is_capture": self.callback_obj_dict.get("is_capture"),
            "is_refunded": self.callback_obj_dict.get("is_refunded"),
            "is_standalone_payment": self.callback_obj_dict.get("is_standalone_payment"),
            "is_voided": self.callback_obj_dict.get("is_voided"),
            "order.id": self.callback_obj_dict.get("order", {}).get("id"),
            "owner": self.callback_obj_dict.get("owner"),
            "pending": self.callback_obj_dict.get("pending"),
            "source_data.pan": self.callback_obj_dict.get("source_data", {}).get("pan"),
            "source_data.sub_type": self.callback_obj_dict.get("source_data", {}).get("sub_type"),
            "source_data.type": self.callback_obj_dict.get("source_data", {}).get("type"),
            "success": self.callback_obj_dict.get("success"),
        }

        return self._generate_processed_hmac(hmac_dict=hmac_dict)

    def _generate_card_token_processed_hmac(self) -> str:
        """Creates HMAC from sent card token callback body_dic

        Returns:
            str: Generated HMAC
        """
        if not isinstance(self.callback_obj_dict, dict):
            return ""

        hmac_dict = {
            "card_subtype": self.callback_obj_dict.get("card_subtype"),
            "created_at": self.callback_obj_dict.get("created_at"),
            "email": self.callback_obj_dict.get("email"),
            "id": self.callback_obj_dict.get("id"),
            "masked_pan": self.callback_obj_dict.get("masked_pan"),
            "merchant_id": self.callback_obj_dict.get("merchant_id"),
            "order_id": self.callback_obj_dict.get("order_id"),
            "token": self.callback_obj_dict.get("token"),
        }

        return self._generate_processed_hmac(hmac_dict=hmac_dict)

    def _generate_delivery_status_processed_hmac(self) -> str:
        """Creates HMAC from sent Delivery Status callback body_dic

        Returns:
            str: Generated HMAC
        """
        if not isinstance(self.callback_obj_dict, dict):
            return ""

        hmac_dict = {
            "order_id": self.callback_obj_dict.get("order_id"),
            "order_delivery_status": self.callback_obj_dict.get("order_delivery_status"),
            "merchant_id": self.callback_obj_dict.get("merchant_id"),
            "merchant_name": self.callback_obj_dict.get("merchant_name"),
            "updated_at": self.callback_obj_dict.get("updated_at"),
        }

        return self._generate_processed_hmac(hmac_dict=hmac_dict)

    # Public Method that can be used Directly to Validate HMAC
    @property
    def is_valid(self) -> bool:
        """Validates HMAC for processed callback

        Returns:
            bool: True if HMAC is Valid, False otherwise
        """
        if not isinstance(self.callback_dict, dict):
            return False

        callback_type = self.callback_dict.get("type")
        if callback_type == AcceptCallbackTypes.TRANSACTION:
            calculated_hmac = self._generate_transaction_processed_hmac()
        elif callback_type == AcceptCallbackTypes.CARD_TOKEN:
            calculated_hmac = self._generate_card_token_processed_hmac()
        elif callback_type == AcceptCallbackTypes.DELIVERY_STATUS:
            calculated_hmac = self._generate_delivery_status_processed_hmac()
        else:
            return False

        if calculated_hmac != self.incoming_hmac:
            return False

        return True



# Python Standard Library Imports
from datetime import datetime

# Other Third Party Imports
from pydantic import BaseModel

from .data import DataDataClass
from .order import OrderDataClass
from .source_data import SourceDataDataClass


class TransactionDataClass(BaseModel):
    id: int
    pending: bool
    amount_cents: int
    success: bool
    is_auth: bool
    is_capture: bool
    is_voided: bool
    is_refunded: bool
    is_standalone_payment: bool
    is_3d_secure: bool
    has_parent_transaction: bool
    integration_id: int
    profile_id: int
    order: OrderDataClass
    currency: str
    source_data: SourceDataDataClass
    data: DataDataClass
    transaction_processed_callback_responses: list
    error_occured: bool
    owner: int
    parent_transaction: bool
    created_at: datetime



# First Party Imports
from paymob.accept.factories import DeliveryStatusCallbackFactory
from paymob.accept.hmac_validator import HMACValidator


class DeliveryStatusCallback(HMACValidator, DeliveryStatusCallbackFactory):
    pass


__all__ = ["DeliveryStatusCallback"]



# Python Standard Library Imports
from typing import Union

# Other Third Party Imports
from pydantic import BaseModel


class OrderItemDataClass(BaseModel):
    name: str
    amount_cents: Union[str, int]
    quantity: Union[str, int]
    description: str



# Python Standard Library Imports
from datetime import datetime
from typing import List, Optional

# Other Third Party Imports
from pydantic import BaseModel

from .order_item import OrderItemDataClass
from .shipping_data import ShippingDataDataClass


class OrderDataClass(BaseModel):
    id: int
    created_at: datetime
    delivery_needed: bool
    amount_cents: int
    currency: str
    is_payment_locked: bool
    merchant_order_id: Optional[str] = None
    wallet_notification: Optional[str] = None
    paid_amount_cents: int
    items: List[OrderItemDataClass]
    shipping_data: ShippingDataDataClass



# Response Codes
SUCCESS = 10

# Request Related Error Codes
JSON_DECODE_EXCEPTION = 20
REQUEST_EXCEPTION = 21
HTTP_EXCEPTION = 22
UNHANDLED_EXCEPTION = 23


# Error Messages Templates
JSON_DECODE_EXCEPTION_MESSAGE = "An Error Occurred While Parsing the Response into JSON"
REQUEST_EXCEPTION_MESSAGE = "An Error Occurred During the Request"
HTTP_EXCEPTION_MESSAGE = "Non 2xx Status Code Returned."
UNHANDLED_EXCEPTION_MESSAGE = "Unhandled Exception"



# Python Standard Library Imports
from dataclasses import dataclass
from typing import Any, Optional


@dataclass
class ResponseFeedBack:

    message: Optional[str]
    data: Any = None
    status_code: int = 500



# Future Imports
from __future__ import annotations

# Python Standard Library Imports
from typing import Tuple, Union

from .config import URLsConfig
from .connection import AcceptConnection
from .data_classes import TransactionDataClass
from .factories import DynamicTransactionFactory
from .response_codes import SUCCESS

# TODO: Allow it to return DynamicTransactionFactory or TransactionDataClass once Validation settings is supported
AbstractTransaction = (
    DynamicTransactionFactory or TransactionDataClass
)  # NOTE: It will always return DynamicTransactionFactory for Now


class Transaction(AbstractTransaction):
    """Final Transaction Class"""

    def __init__(self, connection: AcceptConnection, *args, **kwargs) -> None:
        """Initializing Transaction Attributes

        Args:
            connection (AcceptConnection): Accept Connection Insatance
        """
        super().__init__(*args, **kwargs)
        self.connection = connection

    def __str__(self) -> str:
        if hasattr(self, "id"):
            return f"Transaction No: {self.id}"
        return self.__repr__()

    def refund(
        self,
        amount_cents: int,
    ) -> Tuple[str, Union[Transaction, None], Union[str, None]]:
        """Refund a Transaction

        Args:
            amount_cents (int): Amount that will be Refunded

        Returns:
            Tuple[str, Union[Transaction, None], Union[str, None]]: (Code, Transaction Instance, Success/Error Message)
        """

        request_body = {
            "transaction_id": self.id,
            "amount_cents": amount_cents,
        }

        code, transaction_data, message = self.connection.post(
            url=URLsConfig.REFUND_TRANSACTION,
            json=request_body,
        )

        # TODO: Validates APIs Return Data Option
        transaction_instance = None
        if code == SUCCESS:
            transaction_instance = Transaction(connection=self.connection, **transaction_data)
            message = "Transaction: {0} Refund Processed Successfully".format(
                self.id,
            )
        return code, transaction_instance, message

    def void(
        self,
    ) -> Tuple[str, Union[Transaction, None], Union[str, None]]:
        """Void a Transaction

        Returns:
            Tuple[str, Union[Transaction, None], Union[str, None]]: (Code, Transaction Instance, Success/Error Message)
        """

        request_body = {
            "transaction_id": self.id,
        }

        code, transaction_data, message = self.connection.post(
            url=URLsConfig.VOID_TRANSACTION,
            json=request_body,
        )

        # TODO: Validates APIs Return Data Option
        transaction_instance = None
        if code == SUCCESS:
            transaction_instance = Transaction(connection=self.connection, **transaction_data)
            message = "Transaction: {0} Void Processed Successfully".format(
                self.id,
            )
        return code, transaction_instance, message

    def capture(
        self,
        amount_cents: int,
    ) -> Tuple[str, Union[Transaction, None], Union[str, None]]:
        """Capture a Transaction

        Args:
            amount_cents (int): Amount that will be Captured

        Returns:
            Tuple[str, Union[Transaction, None], Union[str, None]]: (Code, Transaction Instance, Success/Error Message)
        """

        request_body = {
            "transaction_id": self.id,
            "amount_cents": amount_cents,
        }

        code, transaction_data, message = self.connection.post(
            url=URLsConfig.CAPTURE,
            json=request_body,
        )

        # TODO: Validates APIs Return Data Option
        transaction_instance = None
        if code == SUCCESS:
            transaction_instance = Transaction(connection=self.connection, **transaction_data)
            message = "Transaction: {0} Capture Processed Successfully".format(
                self.id,
            )
        return code, transaction_instance, message



# Python Standard Library Imports
from datetime import datetime
from typing import List

# Other Third Party Imports
from pydantic import BaseModel


class MerchantDataClass(BaseModel):
    id: int
    phones: List[str]
    company_emails: List[str]
    created_at: datetime
    company_name: str
    country: str
    state: str
    city: str
    street: str
    postal_code: str



# Python Standard Library Imports
import json
from typing import Any, Dict, Tuple, Union

# Other Third Party Imports
import requests

# First Party Imports
from paymob.data_classes import ResponseFeedBack

from .config import ACCEPT_APIS_TIMEOUT_SECONDES, Credentials, URLsConfig
from .response_codes import (
    HTTP_EXCEPTION,
    HTTP_EXCEPTION_MESSAGE,
    JSON_DECODE_EXCEPTION,
    JSON_DECODE_EXCEPTION_MESSAGE,
    REQUEST_EXCEPTION,
    REQUEST_EXCEPTION_MESSAGE,
    SUCCESS,
    UNHANDLED_EXCEPTION,
    UNHANDLED_EXCEPTION_MESSAGE,
)


class AcceptConnection:
    def __init__(self) -> None:
        """Initializing the Following:
        1- Requests Session
        2- Auth Token
        3- Set Headers
        """
        self.session = requests.Session()
        self.auth_token = self._get_auth_token()
        self.session.headers.update(self._get_headers())

    def _get_headers(self) -> Dict[str, Any]:
        """Initialize Header for Requests

        Returns:
            Dict[str, Any]: Initialized Header Dict
        """
        return {
            "Content-Type": "application/json",
            "Authorization": f"{self.auth_token}",
        }

    def _get_auth_token(self) -> Union[str, None]:
        """Retrieve an Auth Token

        Returns:
            Union[str, None]: Auth Token
        """

        request_body = {"api_key": Credentials.ACCEPT_API_KEY}

        code, data, _ = self.post(
            url=URLsConfig.AUTH_TOKEN,
            json=request_body,
        )

        # TODO: Validates APIs Return Data Option
        token = None
        if code == SUCCESS:
            token = data.get("token")
        return token

    def get(self, *args, **kwargs) -> Tuple[str, Dict[str, Any], ResponseFeedBack]:
        """Wrapper of requests.get method

        Args:
            Same Args of requests.post/requests.get methods

        Returns:
            Tuple[str, Dict[str, Any], ResponseFeedBack]: Tuple containes the Following (Code, Data, Success/Error Message)
        """
        # TODO: The Following Logic will be Abstracted
        reponse_data = None
        try:
            response = self.session.get(timeout=ACCEPT_APIS_TIMEOUT_SECONDES, *args, **kwargs)
            reponse_data = response.json()
            response.raise_for_status()
        except json.JSONDecodeError as error:
            reponse_feedback = ResponseFeedBack(
                message=JSON_DECODE_EXCEPTION_MESSAGE,
                data=error,
                status_code=response.status_code,
            )
            return JSON_DECODE_EXCEPTION, None, reponse_feedback
        except requests.exceptions.HTTPError as error:
            reponse_feedback = ResponseFeedBack(
                message=HTTP_EXCEPTION_MESSAGE,
                data=reponse_data,
                status_code=response.status_code,
            )
            return HTTP_EXCEPTION, reponse_data, reponse_feedback
        except requests.exceptions.RequestException as error:
            reponse_feedback = ResponseFeedBack(message=REQUEST_EXCEPTION_MESSAGE, data=error)
            return REQUEST_EXCEPTION, None, reponse_feedback
        except Exception as error:
            reponse_feedback = ResponseFeedBack(
                message=UNHANDLED_EXCEPTION_MESSAGE,
                data=error,
                status_code=response.status_code,
            )
            return UNHANDLED_EXCEPTION, None, reponse_feedback

        message = "API Successfully Called."
        reponse_feedback = ResponseFeedBack(message=message, status_code=response.status_code)
        return SUCCESS, reponse_data, reponse_feedback

    def post(self, *args, **kwargs) -> Tuple[str, Dict[str, Any], ResponseFeedBack]:
        """Wrapper of requests.get method

        Args:
            Same Args of requests.post/requests.get methods

        Returns:
            Tuple[str, Dict[str, Any], ResponseFeedBack]: Tuple containes the Following (Code, Data, Success/Error Message)
        """
        # TODO: The Following Logic will be Abstracted
        reponse_data = None
        try:
            response = self.session.post(timeout=ACCEPT_APIS_TIMEOUT_SECONDES, *args, **kwargs)
            reponse_data = response.json()
            response.raise_for_status()
        except json.JSONDecodeError as error:
            reponse_feedback = ResponseFeedBack(
                message=JSON_DECODE_EXCEPTION_MESSAGE,
                data=error,
                status_code=response.status_code,
            )
            return JSON_DECODE_EXCEPTION, None, reponse_feedback
        except requests.exceptions.HTTPError as error:
            reponse_feedback = ResponseFeedBack(
                message=HTTP_EXCEPTION_MESSAGE,
                data=reponse_data,
                status_code=response.status_code,
            )
            return HTTP_EXCEPTION, reponse_data, reponse_feedback
        except requests.exceptions.RequestException as error:
            reponse_feedback = ResponseFeedBack(message=REQUEST_EXCEPTION_MESSAGE, data=error)
            return REQUEST_EXCEPTION, None, reponse_feedback
        except Exception as error:
            reponse_feedback = ResponseFeedBack(
                message=UNHANDLED_EXCEPTION_MESSAGE,
                data=error,
                status_code=response.status_code,
            )
            return UNHANDLED_EXCEPTION, None, reponse_feedback

        message = "API Successfully Called."
        reponse_feedback = ResponseFeedBack(message=message, status_code=response.status_code)
        return SUCCESS, reponse_data, reponse_feedback



# Other Third Party Imports
from decouple import config

# Constants
ACCEPT_APIS_TIMEOUT_SECONDES = config("ACCEPT_APIS_TIMEOUT_SECONDES", cast=int, default=10)

# Credentials Config
class Credentials:
    ACCEPT_API_KEY = config("ACCEPT_API_KEY", default="")
    ACCEPT_HMAC_SECRET = config("ACCEPT_HMAC_SECRET", default="")


# URLs Config
class URLsConfig:
    ACCEPT_BASE_URL = config("PAYMOB_ACCEPT_BASE_URL", default="https://accept.paymob.com")
    AUTH_TOKEN = ACCEPT_BASE_URL + "/api/auth/tokens"
    CREATE_ORDER = ACCEPT_BASE_URL + "/api/ecommerce/orders"
    GET_ORDER = ACCEPT_BASE_URL + "/api/ecommerce/orders/{order_id}"
    PAYMENT_KEY = ACCEPT_BASE_URL + "/api/acceptance/payment_keys"
    PAY = ACCEPT_BASE_URL + "/api/acceptance/payments/pay"
    IFRAME = ACCEPT_BASE_URL + "/api/acceptance/iframes/{iframe_id}?payment_token={payment_key}"
    GET_TRANSACTION_ID = ACCEPT_BASE_URL + "/api/acceptance/transactions/{transaction_id}"
    GET_TRANSACTION_MERCHANT_ID = ACCEPT_BASE_URL + "/api/ecommerce/orders/transaction_inquiry"
    REFUND_TRANSACTION = ACCEPT_BASE_URL + "/api/acceptance/void_refund/refund"
    VOID_TRANSACTION = ACCEPT_BASE_URL + "/api/acceptance/void_refund/void"
    CAPTURE = ACCEPT_BASE_URL + "/api/acceptance/capture"
    CREATE_INVOICE_LINK = ACCEPT_BASE_URL + "/api/ecommerce/orders"
    CREATE_PRODUCT_LINK = ACCEPT_BASE_URL + "/api/ecommerce/products"



# First Party Imports
from paymob.accept.constants import AcceptCallbackTypes

from .card_token_callback import CardTokenCallback
from .delivery_status_callback import DeliveryStatusCallback
from .transaction_callback import TransactionCallback


def AcceptCallback(incoming_hmac, callback_dict):

    callback_type = callback_dict.get("type")
    callback_class = None
    if callback_type == AcceptCallbackTypes.TRANSACTION:
        callback_class = TransactionCallback
    elif callback_type == AcceptCallbackTypes.CARD_TOKEN:
        callback_class = CardTokenCallback
    elif callback_type == AcceptCallbackTypes.DELIVERY_STATUS:
        callback_class = DeliveryStatusCallback
    else:
        raise TypeError("Invalid Callback Type")

    return callback_class(incoming_hmac=incoming_hmac, callback_dict=callback_dict, **callback_dict)



import codecs
import os

from setuptools import setup, find_packages

here = os.path.abspath(os.path.dirname(__file__))

with codecs.open(os.path.join(here, "README.md"), encoding="utf-8") as fh:
    long_description = "\\n" + fh.read()

setup(
    name="Paymob-Toolbox",
    version='0.0.1',
    author="DjangoFam Team",
    author_email="mahmoud.nasser.abdulhamed11@gmail.com",
    description="A simple Python package that provides convenient access to the Paymob APIs from applications written in the Python language.",
    url="https://github.com/muhammedattif/Paymob",
    long_description_content_type="text/markdown",
    long_description=long_description,
    packages=find_packages(),
    install_requires=[],
    keywords=['pypi', 'paymob', 'payment', 'python'],
    classifiers=[
        "Development Status :: 1 - Planning",
        "Intended Audience :: Developers",
        "Programming Language :: Python :: 3",
        "Operating System :: Unix",
        "Operating System :: MacOS :: MacOS X",
        "Operating System :: Microsoft :: Windows"
    ]
)



from .card_token_callback import CardTokenCallback
from .delivery_status_callback import DeliveryStatusCallback
from .transaction_callback import TransactionCallback
from .generic_callback import AcceptCallback

__all__ = [AcceptCallback]



# Python Standard Library Imports
from typing import Optional

# Other Third Party Imports
from pydantic import BaseModel


class DataDataClass(BaseModel):
    merchant_txn_ref: str
    card_num: Optional[str] = None
    avs_result_code: str
    merchant: str
    avs_acq_response_code: str
    transaction_no: str
    batch_no: str
    message: str
    txn_response_code: str
    txn_response_code: str
    card_type: str
    receipt_no: str
    created_at: str
    currency: str
    klass: str
    authorize_id: str
    amount: str
    acq_response_code: str
    command: str
    gateway_integration_pk: int
